// Code to run the fablab cottbus decoration lamp
// by Maximilian Voigt, Stefan Arndt

#include <FastLED.h>

#include <array>
#include <functional>

// configuration variables
#define NUMBER_OF_LEDS 3   // How many WS2812 leds do you have?
#define LED_BRIGHTNESS 20  // LED_BRIGHTNESS of your LEDs
#define CONTROL_PIN D0     // trigger pin to connect to the button
#define DATA_PIN D3        // data pin to connect your LEDs
#define BUTTON_PIN D5      // external interrupt pin
#define debounceTime 100  // minimum duration between button presses; in ms

// runtime variables
volatile bool waitingForButtonPress = true; // flag used from inside the ISR
bool blockInterrupts = true;                // inhibit interrupts until initialisation is complete
char lightMode = 0;                         // what mode to start at (0 is off)
std::array<CRGB,NUMBER_OF_LEDS> leds;       // memory for LEDs
CRGBPalette16 currentPalette;               // palette for mode 2

/*
 * #####################################################################################################################
 * LED color functions - adjust to your liking
 * #####################################################################################################################
 */

void switchLedsOff()
{
    leds.fill(0x0);
}

void setLedsToWhite()
{
    leds.fill(CRGB::Red);
}

void pulse()
{
    static CRGB baseColor(0,50,50); // green,red,blue
    leds.fill(baseColor);

    float amp = sin(0.001 * millis()); // each cycle will be 2*pi seconds
    amp = 0.5 * (1 + amp);             // re-scale from [-1,1] to [0,1]
        
    FastLED.setBrightness(amp*255);
}

void simpleWave()
{
    int now = millis();
    for(int i = 0; i < leds.size(); ++i)
    {
        float offset = M_PI/4*i;
        float amp = sin(offset + 0.001 * millis() * M_PI * 10);
        amp = 0.5 * (1 + amp); // re-scale from [-1,1] to [0,1]

        leds[i] = CRGB(amp*255,0,255);
    }
}

void multicolorPulse()
{
    // define the colors you want
    static const std::array<CRGB,4> colors = {
      CRGB(0,255,255),
      CRGB(0,0,0),
      CRGB(0,255,0),
      CRGB(30,30,30)};
    // how long one pulse is; in milliseconds
    const float pulseLength = 200; 

    static uint8_t currentColorIndex = 0; // active color to show
    static int lastColorChange = 0;       // time point of last color change
    
    // shift the sine wave to the right to be in sync with the color change
    const float offset = 1.5f*M_PI;
    
    int now = millis();
    int timeDifference = now - lastColorChange;
    float amp = sin(offset + timeDifference * 2.0 * M_PI / pulseLength);
    amp = 0.5 * (1 + amp); // re-scale from [-1,1] to [0,1]

    if(timeDifference > pulseLength)
    {
        lastColorChange = now;
        ++currentColorIndex;
        if(currentColorIndex == colors.size())
            currentColorIndex = 0;
    }

    leds.fill(colors[currentColorIndex]);

    FastLED.setBrightness(amp*255);
}

void showCustomPalette()
{
    static float index = 0;
    index += 2.0; // adjust this to change the speed

    uint8_t startIndex = (int)index;
    fillLEDsFromPaletteColors(index);
}

void fillLEDsFromPaletteColors(uint8_t colorIndex)
{
    uint8_t brightness = 255;
    TBlendType blending = LINEARBLEND;

    for(CRGB& color : leds)
    {
        color = ColorFromPalette(currentPalette, colorIndex, brightness, blending);
        colorIndex += 3;
    }
}

// available light modes - adjust to your liking
std::array<std::function<void()>,6> lightModes = { switchLedsOff, setLedsToWhite, pulse, simpleWave, multicolorPulse, showCustomPalette };

/*
 * #####################################################################################################################
 * MAIN CODE - DO NOT EDIT
 * #####################################################################################################################
 */

void setup()
{
    Serial.begin(9600);

    pinMode(BUTTON_PIN, INPUT_PULLUP);
    pinMode(CONTROL_PIN, OUTPUT);
    digitalWrite(CONTROL_PIN, LOW);

    // define LEDs
    FastLED.addLeds<WS2812B, DATA_PIN, RGB>(leds.data(), NUMBER_OF_LEDS);
    FastLED.setBrightness(LED_BRIGHTNESS);

    // switch on LEDs shortly to indicate power
    setLedsToWhite();
    delay(1000);
    switchLedsOff();

    currentPalette = RainbowColors_p; // use predefined rainbow as active palette

    attachInterrupt(digitalPinToInterrupt(BUTTON_PIN), buttonInterruptHandler, FALLING);

    blockInterrupts = false; // open interrupt processing for business
}

void loop()
{
    if(hasButtonPress() == true)
    {
        Serial.println("button pressed");

        lightMode++;
        if(lightMode >= lightModes.size())
            lightMode = 0;
    }

    // in case this needs to be reset because one light function changed it
    FastLED.setBrightness(LED_BRIGHTNESS);

    // run active function (yes, slightly wasteful for static modes to run this every time)
    lightModes[lightMode]();

    FastLED.show();
}

// ISR for handling interrupt triggers arising from associated button switch
ICACHE_RAM_ATTR void buttonInterruptHandler()
{
    if(blockInterrupts)
        return;

    if(waitingForButtonPress == true)
    {
        // new interrupt so okay start a new button read process -
        // now need to wait for button release plus debounce period to elapse
        // this will be done in the button_read function
        if(digitalRead(BUTTON_PIN) == LOW)
        {
            // button pressed, so we can start the read on/off + debounce cycle wich
            // will be completed by the button_read() function.
            waitingForButtonPress = false; // keep this ISR 'quiet' until button read fully completed
        }
    }
}

bool hasButtonPress()
{
    if(waitingForButtonPress)
        return false;

    // interrupt has been raised on this button so now need to complete the button
    // read process, ie wait until it has been released and debounce time elapsed

    // static variables because we need to retain old values between function
    // calls
    static bool waitingForRelease = false;
    static long int pressedTimePoint = 0;

    bool isPressed = digitalRead(BUTTON_PIN) == LOW;
    if(isPressed)
    {
        // switch is pressed, so take time and wait for button release
        waitingForRelease = true;
        pressedTimePoint = millis();
    }
    else if(waitingForRelease && (isPressed == false))
    {
        // switch was pressed, now released, so check if debounce time elapsed
        if(millis() - pressedTimePoint >= debounceTime)
        {
            // debounce time elapsed, so the switch press cycle is complete
            waitingForRelease = false;    // reset for next button press interrupt cycle
            waitingForButtonPress = true; // reopen ISR for business now button
                                          // on/off/debounce cycle complete
            return true;                  // return that switch has been pressed
        }
    }

    return false; // either not pressed or debounce period not elapsed
}
